#include <stdio.h>
#include <math.h>
#include <gl.h>
#include <device.h>

char *Progname;
char Usage[] = "usage: %s [-p<partsize>] [file]\n";
#define USAGE() fprintf(stderr, Usage, Progname)

/* Point data. */

typedef struct {
	int x;
	int y;
	int part;       /* which partition this point is in */
} POINT;

#define MAXPOINTS 10000
POINT Data[MAXPOINTS];
int Npoints = 0;

int Minx;
int Miny;
int Maxx;
int Maxy;

/* Partition data. */

typedef struct {
	int l;          /* a rect */
	int r;
	int b;
	int t;
	int n;          /* the number of points in this rect */
	int *p;         /* array of that many points (indices to Data) */
} PART;

#define MAXPARTS 256
PART *Part[MAXPARTS];   /* array of partitions */
int Nparts;

int Partsize;           /* Max # of points in a partition. */

typedef float **MAT;    /* A matrix type. */

MAT Tmat;               /* The transition matrix for the symbol sequence. */
MAT Imat;               /* The incidence matrix. */

void read_data(FILE *);
void do_partition();
void partition(PART *);
PART *new_part(int);
void free_part(PART *);
int find_split(PART *, int);
void label_points();
void symbol_entropy();
void transition_matrix();
float topological_entropy();
float metric_entropy();
MAT new_matrix(int, int);
void free_matrix(MAT, int);
void mat_mul(MAT, MAT, MAT);
void mat_copy(MAT, MAT);

main(int argc, char **argv)
{
	register char *curr_arg;
	register int c;
	FILE *infile;
	float me, te;
	short val;

	Progname = *argv;

	/* Parse arguments beginning with a hyphen. */

	Partsize = 10;
	while (--argc > 0 && **++argv == '-' && *(curr_arg = 1 + *argv)) {

		/* Scan current argument until a flag indicates that the rest
		of the argument isn't flags (curr_arg = NULL), or until
		the end of the argument is reached (if it is all flags). */

		while (curr_arg != NULL && (c = *curr_arg++)) {
			switch (c) {

			case 'p':
				Partsize = atoi(curr_arg);
				curr_arg = NULL;
				break;

			default:
				fprintf(stderr, "%s: unknown flag -%c\n", Progname, c);
				USAGE();
				exit(1);
			}
		}
	}

	/* Process the named file, or stdin if no file given.
	The name '-' also specifies stdin. */

	infile = stdin;
	if (argc && **argv != '-') {
		if ((infile = fopen(*argv, "rt")) == NULL) {
			fprintf(stderr, "%s: can't open '%s'\n", Progname, *argv);
			exit(1);
		}
	}

	/* Read the data in. */

	read_data(infile);

	/* Partition it to create a symbolic dynamics - the sequence of
	partitions the trajectory enters. */

	do_partition();

	/* Create the symbolic orbit.  Label each point according to
	the partition it's in. */

	label_points();

	/* Compute the symbol entropy.  This should be close to the
	max, since we try to make the partition probabilities uniform. */

	symbol_entropy();

	/* Compute the transition matrix and incidence matrix for the
	symbolic orbit. */

	transition_matrix();

	/* Estimate the topological entropy for the symbolic orbit. */

	te = topological_entropy();

	/* Estimate the metric entropy. */

	me = metric_entropy();

	/* Compare the two entropies. */

	printf("%f:", fabs(te - me));
	printf("%f\n", me / te);

#ifdef notdef
	/* Display the partitions. */

	foreground();
	prefsize(Maxx - Minx + 20, Maxy - Miny + 20);
	winopen("");
	ortho2(Minx - 10, Maxx + 10, Miny - 10, Maxy + 10);

	drawpart();

	qdevice(ESCKEY);
	qdevice(REDRAW);
	while (1) {
		c = qread(&val);
		if (c == ESCKEY) {
			break;
		}
		if (c == REDRAW) {
			reshapeviewport();
			drawpart();
		}
	}
#endif
	exit(0);
}

void
read_data(FILE *infile)
{
	int i;
	char buf[100];

	/* Read the data in. */

	i = 0;
	while (fgets(buf, 100, infile)) {
		Data[i].x = atoi(buf);
		fgets(buf, 100, infile);
		Data[i].y = atoi(buf);
		Data[i].part = -1;

#ifdef notdef
		/* Count this point only if it is different from the
		previous point. */

		if (i == 0 || (Data[i].x != Data[i-1].x ||
			       Data[i].y != Data[i-1].y)) {
			i++;
		}
#endif
		i++;

		if (i == MAXPOINTS) {
			fprintf(stderr, "%s: too many points - remainder ignored\n", Progname);
			break;
		}
	}
	Npoints = i;

	/* Get the range. */

	Minx = Miny = 30000;
	Maxx = Maxy = 0;
	for (i = 0; i < Npoints; i++) {
		if (Data[i].x < Minx) {
			Minx = Data[i].x;
		}
		if (Data[i].x > Maxx) {
			Maxx = Data[i].x;
		}
		if (Data[i].y < Miny) {
			Miny = Data[i].y;
		}
		if (Data[i].y > Maxy) {
			Maxy = Data[i].y;
		}
	}
}

void
do_partition()
{
	int i, *p;
	PART *part;

	/* Make the first partition by hand; it is the bounding rect, and
	it contains all the points, in order. */

	part = new_part(Npoints);
	part->l = Minx;
	part->b = Miny;
	part->r = Maxx;
	part->t = Maxy;
	p = part->p;
	for (i = 0; i < Npoints; i++) {
		*p++ = i;
	}
	Nparts = 0;

	/* Recursively subdivide the original partition. */

	partition(part);

	/* Give a summary of the results. */

	printf("%d:", Nparts);
#ifdef notdef
	printf("%d partitions, sizes:\n", Nparts);
	for (i = 0; i < Nparts; i++) {
		printf("%5d", Part[i]->n);
	}
	putchar('\n');
#endif
}

void
partition(PART *part)
{
	int i, n, *p, *pp, xaxis;
	PART *left, *right;

	if (part->n <= Partsize) {
		Part[Nparts++] = part;
		if (Nparts == MAXPARTS) {
			fprintf(stderr, "%s: increase MAXPARTS\n", Progname);
			exit(1);
		}
		return;
	}

	/* Which way to split this partition? */

	xaxis = FALSE;
	if (part->r - part->l > part->t - part->b) {
		xaxis = TRUE;
	}

	/* Sort the values along that axis, and split it into two parts. */

	n = find_split(part, xaxis);

	/* Make two new partitions, and divide the points up. */

	left = new_part(n);
	right = new_part(part->n - n);
	if (xaxis) {
		left->r = right->l = Data[part->p[n]].x;
		left->l = part->l;
		right->r = part->r;
		left->b = right->b = part->b;
		left->t = right->t = part->t;
	} else {
		left->t = right->b = Data[part->p[n]].y;
		left->b = part->b;
		right->t = part->t;
		left->l = right->l = part->l;
		left->r = right->r = part->r;
	}

	/* n is the number of points in the left partition. */

	p = part->p;
	pp = left->p;
	for (i = 0; i < n; i++) {
		*pp++ = *p++;
	}
	pp = right->p;
	for (i = part->n; i > n; --i) {
		*pp++ = *p++;
	}

	free_part(part);
	partition(left);
	partition(right);
}

/* Routines for managing partitions. */

PART *
new_part(int n)
{
	PART *p;
	int *ip;

	p = (PART *)malloc(sizeof(PART));
	ip = (int *)malloc(n * sizeof(int));
	if (p == NULL || ip == NULL) {
		fprintf(stderr, "%s: can't malloc()\n", Progname);
		exit(1);
	}

	p->n = n;
	p->p = ip;

	return (p);
}

void
free_part(PART *p)
{
	free(p->p);
	free(p);
}

/* Calculate where to split the partition.  The xaxis flag says whether to
split it in the x (TRUE) direction or y (FALSE) direction.  Return the index
of the start of the second partition. */

int
xcmp(int *p1, int *p2)
{
	return (Data[*p1].x - Data[*p2].x);
}

int
ycmp(int *p1, int *p2)
{
	return (Data[*p1].y - Data[*p2].y);
}

int
find_split(PART *part, int xaxis)
{
	int n, n1, n2, p1, p2, *p;
	int (*cmp)();

	/* Sort the data into ascending order along the proper axis. */

	cmp = ycmp;
	if (xaxis) {
		cmp = xcmp;
	}
	qsort(p = part->p, n = part->n, sizeof(int), cmp);

	/* If n is odd, the partition is [0,(n-1)/2-1],[(n-1)/2,n-1].
	If n is even, the partition is [0,n/2-1],[n/2,n-1].  This way,
	no matter what, the answer (the index of the start of the second
	partition) is always n >> 1. */

	return (n >> 1);
}

/* Draw the partitions. */

drawpart()
{
	int i;
	POINT *p;
	PART *part;

	color(WHITE);
	clear();
	color(BLACK);

	/* Draw the data. */

	p = Data;
	for (i = 0; i < Npoints; i++) {
		pnt2i(p->x, p->y);
		p++;
	}

	/* Draw the partitions. */

	for (i = 0; i < Nparts; i++) {
		part = Part[i];
		recti(part->l, part->t, part->r, part->b);
		part++;
	}
}

/* Label each data point according to the partition it's in.  This
discretization defines the symbolic orbit. */

void
label_points()
{
	int i, j, *p;

	for (i = 0; i < Nparts; i++) {
		for (j = Part[i]->n, p = Part[i]->p; j > 0; --j, p++) {
			if (Data[*p].part != -1) {
				fprintf(stderr, "point %d labeled twice: %d and %d\n", *p, Data[*p].part, i);
			}
			Data[*p].part = i;
		}
	}
	for (i = 0; i < Npoints; i++) {
		if (Data[i].part == -1) {
			fprintf(stderr, "point not labeled: %d\n", i);
		}
	}
}

/* Run through the symbols, and compute the probability of each.  This
gives the symbol entropy, which is theoretically log(Nparts), if the
partitions are all equiprobable.  Note that the symbol is the partition
number. */

void
symbol_entropy()
{
	int i;
	float *p, pe;

	p = (float *)malloc(Nparts * sizeof(float));
	if (p == NULL) {
		fprintf(stderr, "%s: can't malloc()\n", Progname);
		exit(1);
	}
	for (i = 0; i < Nparts; i++) {
		p[i] = 0.0;
	}
	for (i = 0; i < Npoints; i++) {
		p[Data[i].part] += 1.0;
	}
	pe = 0.0;
	for (i = 0; i < Nparts; i++) {
		p[i] /= Npoints;
		if (p[i] != 0.0) {
			pe += -p[i] * log(p[i]);
		}
	}
	printf("%f:", pe / log(Nparts));

	free(p);
}

/* The transition matrix counts the number of times a symbol transitions to
another symbol (including itself).  The incidence matrix is a boolean matrix
that says whether a particular transition occurs. */

void
transition_matrix()
{
	int i, j, k;

	/* Allocate both matrices and zero them. */

	Tmat = new_matrix(Nparts, Nparts);
	Imat = new_matrix(Nparts, Nparts);

	/* Pass thru the data counting all the symbolic transitions. */

	i = Data[0].part;
	for (k = 1; k < Npoints; k++) {
		j = Data[k].part;
		Tmat[i][j] += 1.0;
		i = j;
	}

	/* Mark the transitions that actually occur. */

	for (i = 0; i < Nparts; i++) {
		for (j = 0; j < Nparts; j++) {
			if (Tmat[i][j] != 0.0) {
				Imat[i][j] = 1.0;
			}
		}
	}
}

/* Exponentiate Imat and watch the growth rate of the trace, an estimate of
the largest eigenvalue of Imat, the log of which is the topological entropy. */

float
topological_entropy()
{
	int i, j;
	MAT m, m1;
	float tr, lasttr, gr, lastgr;

	/* Temp matrices.  Start m off as the identity matrix. */

	m = new_matrix(Nparts, Nparts);
	m1 = new_matrix(Nparts, Nparts);

	for (i = 0; i < Nparts; i++) {
		m[i][i] = 1.0;
	}

	lasttr = 1.0;
	lastgr = 0.0;
	for (j = 0; j < 50; j++) {

		/* Multiply m by Imat, and get the trace. */

		mat_mul(m1, m, Imat);
		mat_copy(m, m1);
		tr = 0.0;
		for (i = 0; i < Nparts; i++) {
			tr += m[i][i];
		}

		/* By the Perron-Frobineus theorem, the growth rate
		of the trace converges.  Stop when it does. */

		gr = tr / lasttr;
		if (fabs(gr - lastgr) < 0.001) {
			break;
		}
		lasttr = tr;
		lastgr = gr;
	}
	if (j == 50) {
		printf("[trace convergence cutoff]\n");
	}

	free_matrix(m, Nparts);
	free_matrix(m1, Nparts);

	printf("%f:", log(gr));
	return (log(gr));
}

float
metric_entropy()
{
	int i, j, done;
	MAT m, m0, m1;
	float x, me;

	/* Convert Tmat to a markov probability matrix, m.  Each row of p sums
	to 1, each entry of the row is the probability of transitioning from
	symbol i to symbol j. */

	m = new_matrix(Nparts, Nparts);
	m0 = new_matrix(Nparts, Nparts);
	m1 = new_matrix(Nparts, Nparts);

	for (i = 0; i < Nparts; i++) {

		/* Get the total for this row. */

		x = 0.0;
		for (j = 0; j < Nparts; j++) {
			x += Tmat[i][j];
		}

		/* Convert the row to probabilities. */

		for (j = 0; j < Nparts; j++) {
			m[i][j] = Tmat[i][j] / x;
		}
	}
	mat_copy(m0, m);

#ifdef notdef
	/* Exponentiate the matrix m by successive squarings.  Iterating
	m[n+1] = m[n] * m[n] leads to a fixed point solution, where the rows
	of m are the probabilities of the symbols, and all rows are equal.
    @@@ Maybe this should be 1 / Nparts, instead of this estimate. */

	done = FALSE;
	while (!done) {
		mat_mul(m1, m, m);
		done = TRUE;
		for (i = 0; i < Nparts; i++) {
			if (fabs(m1[i][i] - m[i][i]) >= 0.001) {
				done = FALSE;
				break;
			}
		}
		mat_copy(m, m1);
	}
#endif

	/* Compute the metric entropy from the symbol probabilities in the
	diagonal of m and the original transition probabilities in the
	matrix m0. */

	me = 0.0;
	for (i = 0; i < Nparts; i++) {

		/* Compute x, the entropy of transitioning out of symbol i. */

		x = 0.0;
		for (j = 0; j < Nparts; j++) {
			if (m0[i][j] != 0.0) {
				x += -m0[i][j] * log(m0[i][j]);
			}
		}

		/* Times the probability of being in partition i. */

		me += x / Nparts;
#ifdef notdef
		me += m[i][i] * x;
#endif
	}

	free_matrix(m, Nparts);
	free_matrix(m0, Nparts);
	free_matrix(m1, Nparts);

	printf("%f:", me);
	return (me);
}

/* Create an imax by jmax matrix and zero it. */

MAT
new_matrix(int imax, int jmax)
{
	float **m;
	int i, j;

	m = (float **)malloc(imax * sizeof(float *));
	if (m == NULL) {
		fprintf(stderr, "%s: can't malloc()\n", Progname);
		exit(1);
	}
	for (i = 0; i < imax; i++) {
		m[i] = (float *)malloc(jmax * sizeof(float));
		if (m[i] == NULL) {
			fprintf(stderr, "%s: can't malloc()\n", Progname);
			exit(1);
		}
		for (j = 0; j < jmax; j++) {
			m[i][j] = 0.0;
		}
	}

	return (m);
}

void
free_matrix(MAT m, int imax)
{
	int i;

	for (i = 0; i < imax; i++) {
		free(m[i]);
	}
	free(m);
}

/* Matrix multiplication and copying. @@@ MATs must all be Nparts by Nparts. */

void
mat_mul(MAT d, MAT a, MAT b)
{
	int i, j, k;

	for (i = 0; i < Nparts; i++) {
		for (j = 0; j < Nparts; j++) {
			d[i][j] = 0.0;
			for (k = 0; k < Nparts; k++) {
				d[i][j] += a[i][k] * b[k][j];
			}
		}
	}
}

void
mat_copy(MAT d, MAT s)
{
	int i, j;

	for (i = 0; i < Nparts; i++) {
		for (j = 0; j < Nparts; j++) {
			d[i][j] = s[i][j];
		}
	}
}
